#define _GNU_SOURCE

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <grp.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <syslog.h>
#include <termios.h>
#include <unistd.h>

#include "cgicomm.h"
#include "connectionmgr.h"
#include "fifo.h"
#include "shellinaboxd.h"


static int getNetworkInt(FIFO *fifo)
{
    char buf[4];
    int  offset,i;
    char *chunk;

    for (offset = 0; offset < 4; offset += i ) {
        chunk = fifoGetFirstChunk(fifo,&i);
        if (chunk == NULL)
            return(0);
        if (i > 4-offset)
            i = 4-offset;
        memmove(buf+offset,chunk,i);
        fifoDequeueChunk(fifo,i); }
    return(((int)((unsigned char *)buf)[0] << 24) |
           ((int)((unsigned char *)buf)[1] << 16) |
           ((int)((unsigned char *)buf)[2] <<  8) |
            (int)((unsigned char *)buf)[3]);
}


static void writeNetworkInt(FIFO *fifo,int i)
{
    unsigned char *chunk;

    chunk    = fifoCreateChunk(4);
    chunk[0] = (unsigned char)(i >> 24);
    chunk[1] = (unsigned char)(i >> 16);
    chunk[2] = (unsigned char)(i >>  8);
    chunk[3] = (unsigned char)(i      );
    fifoAppendChunk(fifo,chunk,4);
    return;
}


static int forkPty(int *fd,char *name,int width,int height)
{
    int master,slave,pid;
#ifdef USELOGIN

    // Try to let the standard C library to open a pty pair for us
    master = getpt();
    if (master >= 0) {
        grantpt(master);
        unlockpt(master);
        strcpy(name,ptsname(master));
        if (!strncmp(name,"/dev/tty",8)) {
            // If this is an old-style pty, we need to modify the permissions
            struct group *gr;
            int          ttygid;
            if ((gr = getgrnam("tty")) != NULL)
                ttygid = gr->gr_gid;
            else
                ttygid = -1;
            setreuid(-1,euid);
            chown(name,uid,ttygid);
            chmod(name,S_IRUSR|S_IWUSR|S_IWGRP);
            setreuid(-1,uid);
            slave = open(name,O_RDWR|O_NOCTTY);
            if (slave < 0) {
                setreuid(-1,euid);
                chmod(name,0666);
                chown(name,0,ttygid);
                setreuid(-1,uid);
                close(master);
                return(-1); } }
        else {
            slave = open(name,O_RDWR|O_NOCTTY);
            if (slave < 0) {
                close(master);
                return(-1); } } }
#ifdef USEOLDPTYS
    // Open an old-style pty by sequentially trying all possible entries in
    // /dev/[pt]ty*
    else {
        char         fname[PTYNAMELEN];
        char         *ptr1,*ptr2;
        int          ttygid;
        struct group *gr;
        if ((gr = getgrnam("tty")) != NULL)
            ttygid = gr->gr_gid;
        else
            ttygid = -1;
        strcpy(fname,"/dev/ptyXX");
        master = slave = -1;
        for (ptr1 = "pqrstuvwxyzabcde"; *ptr1; ptr1++) {
            fname[8] = *ptr1;
            for (ptr2 = "0123456789abcdef"; *ptr2; ptr2++) {
                fname[9] = *ptr2;
                if ((master = open(fname,O_RDWR,0)) < 0) {
                    if (errno == ENOENT) {
                        return(-1); } }
                else {
                    fname[5] = 't';
                    // Old-style ptys require updating of permissions
                    setreuid(-1,euid);
                    chown(fname,uid,ttygid);
                    chmod(fname,S_IRUSR|S_IWUSR|S_IWGRP);
                    setreuid(-1,uid);
                    if ((slave = open(fname,O_RDWR|O_NOCTTY)) >= 0) {
                        strcpy(name,fname);
                        goto found_pty; }
                    else {
                        setreuid(-1,euid);
                        chmod(fname,0666);
                        chown(fname,0,ttygid);
                        close(master);
                        setreuid(-1,uid);
                        fname[5] = 'p'; } } } } }
 found_pty:
#endif
    if (master < 0) {
        errno = ENOENT; /* out of ptys */
        return(-1); }
    // Set new window size
    if (width > 0 && height > 0) {
        struct winsize win;
        win.ws_row    = width;
        win.ws_col    = height;
        win.ws_xpixel = 0;
        win.ws_ypixel = 0;
        ioctl(slave,TIOCSWINSZ,&win); }
#else
    int fds[2];
    if (socketpair(PF_UNIX,SOCK_STREAM,0,fds)) {
        return(-1); }
    master = fds[0];
    slave  = fds[1];
    *name  = '\000';
#endif
    // Now, fork off the child process
    if ((pid = fork()) < 0) {
        close(slave);
        close(master);
        return(-1); }
    else if (pid == 0) {
        int i;
        // Close all file handles
        closelog();
        for (i = sysconf(_SC_OPEN_MAX); --i > 0;)
            if (i != slave)
                close(i);
        // Become the session/process-group leader
        setsid();
        setpgid(0,0);
        // Redirect standard I/O to the pty
        dup2(slave,0);
        dup2(slave,1);
        dup2(slave,2);
        if (slave > 2)
            close(slave);
        // Force the pty to be our control terminal
        close(open(name,O_RDWR));
        return(0); }
    fcntl(master,F_SETFL,(long)(O_NONBLOCK|O_RDWR));
    *fd = master;
    close(slave);
    return(pid);
}


#ifndef USELOGIN
static char *readLine(void)
{
    char *buf    = malloc(BLOCKSIZE);
    int  bufSize = BLOCKSIZE;
    int  esc     = 0;
    int  len     = 0;

    for (*buf = '\000'; len < BLOCKSIZE-1; ) {
        int  i;
        if (len >= bufSize-2)
            buf = realloc(buf,bufSize += BLOCKSIZE);
        if ((i = read(0,buf+len,1)) <= 0) {
        eof:
            free(buf);
            return(NULL); }
        buf[++len] = '\000';
        if (esc) {
            esc = 0;
            goto noCtrl; }
        switch (buf[len-1]) {
        case '\000':
        case '\003': // CTRL-C
        case '\004': // CTRL-D
        case '\034': // CTRL-/
            buf[--len] = '\000';
            goto eof;
        case '\010': // CTRL-H
        case '\177': // DEL
            buf[--len] = '\000';
            if (len >= 1) {
                if ((((int)buf[--len])&0xFF) < 32)
                    write(1,"\010 \010",3);
                buf[len] = '\000';
                write(1,"\010 \010",3); }
            break;
        case '\025': // CTRL-U
            buf[--len] = '\000';
            while (len >= 1) {
                if ((((int)buf[--len])&0xFF) < 32)
                    write(1,"\010 \010",3);
                buf[len] = '\000';
                write(1,"\010 \010",3); }
            break;
        case '\022': // CTRL-R
            buf[--len] = '\000';
            write(1,"^R\r\n",4);
            write(1,buf,len);
            break;
        case '\027': // CTRL-W
            {int w = 0;
            buf[--len] = '\000';
            while (len >= 1 && ((w |= !isspace(buf[len-1])) || !w)) {
                if ((((int)buf[--len])&0xFF) < 32)
                    write(1,"\010 \010",3);
                buf[len] = '\000';
                write(1,"\010 \010",3); }
            break;}
        case '\026': // CTRL-V
            buf[--len] = '\000';
            esc++;
            break;
        case '\012': // CTRL-J
        case '\015': // CTRL-M
            buf[--len] = '\000';
            goto done;
        default:
        noCtrl:
            if ((((int)buf[len-1])&0xFF) < 32) {
                char ch[2];
                ch[0] = '^';
                ch[1] = buf[len-1] | '\100';
                write(1,ch,2); }
            else
                write(1,buf+len-1,1);
            break; } }
    done:
    write(1,"\r\n",2);
    buf = realloc(buf,strlen(buf) + 1);
    return(buf);
}
#endif


static int createNewSession(FIFO *fifo,char *terminalType,char eraseChar,
                            struct winsize *winSize,int *sequenceNumber)
{
    int          pid;
    int          pty;
    char         ptyName[PTYNAMELEN];
#ifdef USELOGIN
    struct utmpx utmpx;
#endif

    // Open a new pty and fork off a child process
    pid = forkPty(&pty,ptyName,0,0);
    if (pid < 0) {
        fAbortCGI(fifo,500,"Internal Error\r\n"
                           "Failed to open PTY!");
        return(-1); }
    // Initialize the utmpx structure
#ifdef USELOGIN
    memset(&utmpx,0,sizeof(utmpx));
    utmpx.ut_type  = LOGIN_PROCESS;
    utmpx.ut_pid   = pid ? pid : getpid();
    strncpy(utmpx.ut_line,ptyName+5,      sizeof(utmpx.ut_line));
    strncpy(utmpx.ut_id,  utmpx.ut_line+3,sizeof(utmpx.ut_id));
    strncpy(utmpx.ut_user,"SHELLINABOX",  sizeof(utmpx.ut_user));
    time(&utmpx.ut_tv.tv_sec);
#endif
    if (pid == 0) {
        // Child process:
        static const   char *legalEnv[] = { "TZ", "HZ", NULL };
        char           **env            = calloc(sizeof(char *),4);
        int            numEnvVars       = 0;
        int            i;
        // Clean the environment variables
        if (terminalType != NULL && strlen(terminalType) > 0)
            env[numEnvVars++] = strcat(strcpy(malloc(strlen(terminalType) + 6),
                                              "TERM="),terminalType);
        else
            env[numEnvVars++] = "TERM="TERMINAL;
        sprintf(env[numEnvVars++] = malloc(80),"LINES=%d",winSize->ws_row);
        sprintf(env[numEnvVars++] = malloc(80),"COLUMNS=%d",winSize->ws_col);
        for (i = 0; legalEnv[i] != NULL; i++) {
            char *e = getenv(legalEnv[i]);
            if (e != NULL) {
                env = realloc(env,sizeof(char *)*(1 + ++numEnvVars));
                env[numEnvVars-1] = strcat(strcat(strcpy(malloc(strlen(
                                legalEnv[i]+strlen(e)+2)),legalEnv[i]),"="),e);
                env[numEnvVars  ] = NULL; } }
#ifdef USELOGIN
        {// Set initial terminal settings
	struct termios tt;
	tcgetattr(pty,&tt);
        cfsetispeed(&tt,38400);
        cfsetospeed(&tt,38400);
	tt.c_iflag = TTYDEF_IFLAG & ~ISTRIP;
	tt.c_oflag = TTYDEF_OFLAG;
	tt.c_lflag = TTYDEF_LFLAG;
        tt.c_cflag =(TTYDEF_CFLAG & ~(CS7|PARENB|HUPCL)) | CS8;
        tt.c_cc[VERASE] = eraseChar;
	tcsetattr(pty,TCSAFLUSH,&tt);
	ioctl(pty,TIOCSWINSZ,winSize);
        // Assert privileges
        setreuid(-1,euid);
        // Update utmp entry
        setutxent();
        if (pututxline(&utmpx) == NULL) {
            fprintf(stderr,"pututxline() failed\n");
            exit(1); }
        endutxent();
        // Launch login
        execle(LOGINPATH,"login","-p",NULL,env);
        fprintf(stderr,"Could not execute \"%s\"\n",LOGINPATH); }
#else
        {char *ptr;
        char  *userName;
#ifdef QUERYHOST
        char  *hostName;
      readHostName:
        write(1,"Host: ",6);
        hostName = readLine();
        if (hostName == NULL) {
            fprintf(stderr,"Connection closed by foreign host.\n");
            _exit(1); }
        if (!*hostName || *hostName == '-') {
            free(hostName);
            goto readHostName; }
        for (ptr = hostName; *ptr; ptr++)
            if ((*ptr < '-'              ) ||
                (*ptr > '.' && *ptr < '0') ||
                (*ptr > '9' && *ptr < 'A') ||
                (*ptr > 'Z' && *ptr < 'a') ||
                (*ptr > 'z'              )) {
                free(hostName);
                goto readHostName; }
#else
        char hostName[BLOCKSIZE];
        // Rlogin doesn't ask for the user name, so we need to fake the prompt
        if (!gethostname(hostName,BLOCKSIZE))
            strcpy(hostName,"localhost");
#endif
      readUserName:
        if (*hostName) {
            write(1,hostName,strlen(hostName));
            write(1," ",1); }
        write(1,"login: ",7);
        userName = readLine();
        if (userName == NULL) {
            fprintf(stderr,"Connection closed by foreign host.\n");
            _exit(1); }
        if (!*userName || *userName == '-') {
            free(userName);
            goto readUserName; }
        for (ptr = userName; *ptr; ptr++)
            if (*ptr <= ' ' || *ptr&0x80) {
                free(userName);
                goto readUserName; }
        execle(RLOGINPATH,"rlogin","-8El",userName,hostName,NULL,env);
        fprintf(stderr,"Could not execute \"%s\"\n",RLOGINPATH); }
#endif
        _exit(1); }
    // Create datastructures to keep track of new process
#ifdef USELOGIN
    sessionCreate(pid,pty,ptyName,sequenceNumber,&utmpx);
#else
    sessionCreate(pid,pty,ptyName,sequenceNumber,0);
#endif
    return(pty);
}


static void inputAvailableCallBack(void *data)
{
    ServerData *serverData;
    Connection *connection = findConnection((int)data,&serverData);

    if (connection == NULL)
        return;
    enablePollEvents(connection,POLLOUT);
    return;
}


static void inputEmptyCallBack(void *data)
{
    ServerData *serverData;
    Connection *connection = findConnection((int)data,&serverData);

    if (connection == NULL)
        return;
    disablePollEvents(connection,POLLOUT);
    return;
}


static void inputFlushAndCloseCallBack(void *data)
{
    ServerData *serverData;
    Connection *connection = findConnection((int)data,&serverData);

    if (connection == NULL)
        return;
    enablePollEvents(connection,POLLOUT);
    serverData->u.cg.done = 1;
    return;
}


Connection *cgiCommCreate(int fd)
{
    extern Fnc cgiCommVTable[];
    ServerData *serverData;
    Connection *connection;

    serverData                      = malloc(sizeof(ServerData));
    serverData->vTable              = cgiCommVTable;
    serverData->u.cg.fd             = fd;
    serverData->u.cg.contentLength  = -1;
    serverData->u.cg.outputFifoSize = 0;
    serverData->u.cg.done           = 0;
    serverData->u.cg.inputFifo      = createNewFifo(0,(void *)fd,
                                                   inputAvailableCallBack,
                                                   inputEmptyCallBack,
                                                   NULL,NULL,
                                                   inputFlushAndCloseCallBack);
    serverData->u.cg.outputFifo     = createNewFifo(0,NULL,NULL,NULL,NULL,
                                                    NULL,NULL);
    connection                      = addConnection(fd,serverData);
    enablePollEvents(connection,POLLIN);
    return(connection);
}


static void cgiCommPollIn(int fd,Connection *connection,CGIComm *cgiComm)
{
    char *chunk;
    int  i,count;

    // More data is available from the CGI
    for (;;) {
        if (cgiComm->contentLength < 0) {
            // The first four bytes contain the content length in host order;
            // as long as we don't know the exact length, try reading a large
            // block
            count = BLOCKSIZE; }
        else
            count = cgiComm->contentLength - cgiComm->outputFifoSize;
        if (count <= 0)
            break;
        chunk = fifoCreateChunk(count);
        i     = read(fd,chunk,count);
        if (i <= 0) {
            if (i == 0 || errno == EAGAIN) {
                fifoAppendChunk(cgiComm->outputFifo,chunk,0);
                break; }
            fifoAppendChunk(cgiComm->outputFifo,chunk,0);
            syslog(LOG_DAEMON|LOG_NOTICE,
                   "Read from CGI failed (errno=%d)",errno);
            markConnectionForDeletion(fd);
            return; }
        fifoAppendChunk(cgiComm->outputFifo,chunk,i);
        cgiComm->outputFifoSize += i;
        if (cgiComm->contentLength < 0) {
            if (cgiComm->outputFifoSize >= 4) {
                // We just succeeded in reading the content length; copy it
                // into our structure. We might have to do multiple passes
                // if the FIFO returns multiple chunks.
                int offset,j;
                for (offset = 0; offset < 4; offset += j ) {
                    chunk = fifoGetFirstChunk(cgiComm->outputFifo,&j);
                    if (j > 4-offset)
                        j = 4-offset;
                    memmove((char *)&cgiComm->contentLength+offset,chunk,j);
                    fifoDequeueChunk(cgiComm->outputFifo,j); }
                cgiComm->outputFifoSize -= 4;
                if (cgiComm->contentLength < 8) {
                    syslog(LOG_DAEMON|LOG_NOTICE,
                           "Protocol violation; content length too small");
                    markConnectionForDeletion(fd);
                    return; } } } }
    if (cgiComm->contentLength >= 0 &&
        cgiComm->outputFifoSize >= cgiComm->contentLength) {
        // Read everything
        int  sessionId,sequenceNumber;
        disablePollEvents(connection,POLLIN);
        // The next two integers identify the session;
        // they are transmitted in network order
        sessionId      = getNetworkInt(cgiComm->outputFifo);
        sequenceNumber = getNetworkInt(cgiComm->outputFifo);
        // CONNECT requests are special, because they don't have
        // a session assigned yet
        if (!fifoStrCmp(cgiComm->outputFifo,CONNECT) &&
            cgiComm->contentLength > (strlen(CONNECT) + 6*4 + 1)) {
            static const char *hdr = "Content-type: "WLCONTENTTYPE"\r\n"
                                     "\r\n"CONNECT;
            int            sequenceNumber;
            struct winsize winSize;
            char           eraseChar;
            char           *chunk,*terminalType;
            int            i,offset,len;
            int            id;
            fifoDequeueChunk(cgiComm->outputFifo,strlen(CONNECT));
            winSize.ws_col    = getNetworkInt(cgiComm->outputFifo);
            winSize.ws_row    = getNetworkInt(cgiComm->outputFifo);
            winSize.ws_xpixel = getNetworkInt(cgiComm->outputFifo);
            winSize.ws_ypixel = getNetworkInt(cgiComm->outputFifo);
            chunk             = fifoGetFirstChunk(cgiComm->outputFifo,&i);
            eraseChar         = *chunk;
            fifoDequeueChunk(cgiComm->outputFifo,1);
            len               = cgiComm->contentLength-strlen(CONNECT)-6*4-1;
            terminalType      = calloc(len+1,1);
            for (offset = 0; offset < len; ) {
                chunk = fifoGetFirstChunk(cgiComm->outputFifo,&i);
                if (chunk == NULL)
                    break;
                if (memchr(chunk,'\000',i) != 0) {
                    i   = strlen(chunk);
                    len = offset + i; }
                if (offset + i > len)
                    i = len - offset;
                memmove(terminalType + offset,chunk,i);
                fifoDequeueChunk(cgiComm->outputFifo,i);
                offset += i; }
            id = createNewSession(cgiComm->inputFifo,terminalType,eraseChar,
                                  &winSize,&sequenceNumber);
            free(terminalType);
#ifdef DEBUGREQUESTS
            syslog(LOG_DEBUG|LOG_DAEMON,"CONNECT: %d, 0",id);
#endif
            if (id < 0)
                return;
            chunk = fifoCreateChunk(strlen(hdr));
            memmove(chunk,hdr,strlen(hdr));
            fifoAppendChunk(cgiComm->inputFifo,chunk,strlen(hdr));
            writeNetworkInt(cgiComm->inputFifo,id);
            writeNetworkInt(cgiComm->inputFifo,sequenceNumber);
            fifoFlushAndClose(cgiComm->inputFifo); }
        else { // Everything else is handled by the session
            ServerData *serverData;
            connection = findConnection(sessionId,&serverData);
            if (connection == NULL) {
                fAbortCGI(cgiComm->inputFifo,400,"Illegal Request\r\n"
                                                 "No such session (%d,%d)!",
                          sessionId,sequenceNumber); }
            else
                FNC(serverData,ServeRequest)(sessionId,connection,
                                             &serverData->u,
                                             cgiComm->inputFifo,
                                             sequenceNumber,
                                             cgiComm->outputFifo); } }
    return;
}


static void cgiCommPollOut(int fd,Connection *connection,CGIComm *cgiComm)
{
    // The CGI is willing to receive data
    int  i;

    while ((i = fifoWriteV(fd,cgiComm->inputFifo)) > 0) {
#ifdef DEBUGREPLIES
        {char *buf = malloc(i+1);
        char  *chunk,*dst = buf;
        int   j = 0,k = 0,l;
        
        chunk = fifoGetFirstChunk(cgiComm->inputFifo,&l);
        while (j < i) {
            if (k >= l) {
                chunk = fifoGetNextChunk(chunk,&l); k = 0; }
            *dst++ = chunk[k] <= 32 || chunk[k] >= 127 ? ' ' : chunk[k];
            k++; j++; }
        *dst++ = '\000';
        syslog(LOG_DEBUG|LOG_DAEMON,"%d (%d): \"%s\"",fd,i,buf);
        free(buf); }
#endif
        fifoDequeueChunk(cgiComm->inputFifo,i); }
    if (i < 0) {
        if (errno == EAGAIN)
            return;
        // Write to PTY failed
        syslog(LOG_DAEMON|LOG_NOTICE,
               "Write to CGI failed (errno=%d)",errno);
        markConnectionForDeletion(fd);
        return; }
    if (fifoGetFirstChunk(cgiComm->inputFifo,&i) == NULL &&
        cgiComm->done)
        markConnectionForDeletion(fd);
    return;
}


static void cgiCommPollHup(int fd,Connection *connection,CGIComm *cgiComm)
{
    // Something terrible has happened...
    syslog(LOG_DAEMON|LOG_NOTICE,"Remote end of CGI has been closed");
    markConnectionForDeletion(fd);
    return;
}


static void cgiCommDestroy(int fd,Connection *connection,CGIComm *cgiComm)
{
    // Close socket
#ifdef DEBUGREPLIES
    syslog(LOG_DEBUG|LOG_DAEMON,"Closing connection to CGI: %d",fd);
#endif
    close(fd);
    // Remove old fifos
    destroyFifo(cgiComm->inputFifo);
    destroyFifo(cgiComm->outputFifo);
    return;
}


Fnc cgiCommVTable[VTABSIZE] = {
    (Fnc)cgiCommPollIn,(Fnc)cgiCommPollOut,(Fnc)cgiCommPollHup,
    (Fnc)cgiCommDestroy,NULL,NULL,NULL };

/*
 * Local variables:
 * mode: c
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
