package com.shellinabox;

import com.shellinabox.clipboard.Clipboard;
import com.shellinabox.console.Console;
import com.shellinabox.console.VectorFontConsole;
import com.shellinabox.dialogs.ConfigurationDialog;
import com.shellinabox.emulation.VT102;
import com.shellinabox.persistent.Configuration;
import com.shellinabox.tunnel.GuiObject;
import com.shellinabox.tunnel.Tunnel;
import com.shellinabox.tunnel.TunnelException;
import java.applet.Applet;
import java.applet.AppletContext;
import java.applet.AudioClip;
import java.awt.Button;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Event;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Panel;
import java.awt.Scrollbar;
import java.awt.TextField;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.NoSuchElementException;


public class ShellInABox extends Applet implements GuiObject {

    static final String VERSION    = "v1.0\u03B2 (24Jun01)"; // --- VERSION ---

    private static final int PANEL_SEP       = 1;
    private static final int HEIGHT_FUDGE    = 1;
    private static final int SCROLLBAR_WIDTH = 15;
    private static final int BUTTON_HEIGHT   = 30;
    private static final int BUTTON_WIDTH    = 100;
    private static final int BUTTON_SEP      = 5;
    private static final String CONNECT      = "Connect";
    private static final String DISCONNECT   = "Disconnect";

    private Console             outputArea;
    private TextField           inputField;
    private Button              connectionButton;
    private int                 scrollBackLines;
    private URL                 cgiURL;
    private URL                 fontURL;
    private Color               bgColor;
    private int                 width;
    private int                 height;
    private int                 scrollbarWidth;
    private Configuration       configuration;
    private String              terminalType = "linux";
    private int                 cursorKeyType;
    private boolean             backspaceType;
    private int                 bellType;
    private boolean             nativeClipboard;
    private Clipboard           clipboard;
    private boolean             videoMode;

    boolean                     consoleIsFocused;
    AudioClip                   beepAudioClip;
    VT102                       outputEmulator;
    ConfigurationDialog         configurationDialog;
    Tunnel                      tunnel;

    
    public void init() {
        if (!getParameters())
            return;
        initSimpleGUI();
        return;
    }


    public void destroy() {
        try {
            if (tunnel.isConnected())
                tunnel.disconnect();
        } catch (Exception e) {
            // don't care
        }
        return;
    }


    public void showErrorMessage(String msg) {
        try {
            if (tunnel.isConnected())
                tunnel.disconnect();
        } catch (Exception e) {
                // don't care
        }
        showStatus(msg);
        if (outputArea != null) {
            outputArea.clearScrollBackBuffer();
            outputEmulator.reset();
            outputEmulator.writeString("\007\033[31m" + msg + "\033[m");
        }
        if (inputField != null)
            inputField.setText("");
        return;
    }


    public void receiveDataEvent(String data)
    {
        String response = outputEmulator.writeString(data);
        if (response != null)
            try {
                tunnel.submitText(response);
            } catch (TunnelException e) {
                showErrorMessage(e.getMessage());
            } catch (NullPointerException e) {
                // hmmm
            }
        return;
    }


    public void disconnectEvent() {
        connectionButton.setLabel(CONNECT);
        outputEmulator.writeString("\r\n\r\nConnection closed...");
        showStatus("Connection closed");
        inputField.setText("");
        tunnel = null;
        return;
    }
    

    private boolean getParameters() {
        cgiURL  = getURL("cgi","");
        if (cgiURL == null) {
            showErrorMessage("No valid \"cgi\" provided");
            return(false);
        }
        fontURL = getURL("font","shellinabox.fnt");
        if (fontURL == null) {
            showErrorMessage("No valid \"font\" provided");
            return(false);
        }
        String hiddenFrame = getParameter("hiddenframe");
        if (hiddenFrame == null)
            hiddenFrame = "hidden";
        else if (hiddenFrame.length() == 0)
            hiddenFrame = null;
        if (hiddenFrame != null) {
            URL cookieURL = getURL("cookieurl","blank.html");
            if (cookieURL != null) {
                configuration = new Configuration("ShellInABox",
                                                  getAppletContext(),
                                                  cookieURL,hiddenFrame);
            }
        }
        configurationDialog = new ConfigurationDialog(this,configuration);
        String beepString = getParameter("bell");
        bellType = configurationDialog.getSetting(
                                            ConfigurationDialog.BELL_TYPE_KEY);
        if (bellType == ConfigurationDialog.UNKNOWN ||
            bellType == ConfigurationDialog.FANCY) {
            if (bellType == ConfigurationDialog.FANCY ||
                beepString == null ||
                !beepString.equalsIgnoreCase("visual")) {
                URL beepURL = getURL("bell","beep.au");
                if (beepURL != null) {
                    beepAudioClip = getAudioClip(beepURL);
                    bellType = ConfigurationDialog.FANCY;
                } else
                    bellType = ConfigurationDialog.SYSTEM;
            } else
                bellType = ConfigurationDialog.VISUAL;
            configurationDialog.changeSetting(
                                             ConfigurationDialog.BELL_TYPE_KEY,
                                             bellType, false);
        }
        scrollBackLines = Console.DEFAULT_SCROLLBACK;
        try {
            String scrollBackLinesString = getParameter("scrollback");
            if (scrollBackLinesString != null)
                if (scrollBackLinesString.equals(""))
                    scrollBackLines = 0;
                else
                    scrollBackLines=Integer.parseInt(scrollBackLinesString,10);
            if (scrollBackLines < 0)
                scrollBackLines = 0;
            else if (scrollBackLines > Console.MAXIMUM_SCROLLBACK)
                scrollBackLines = Console.MAXIMUM_SCROLLBACK;
        } catch (NumberFormatException e) {
            // don't care
        }
        scrollbarWidth = SCROLLBAR_WIDTH;
        try {
            String scrollbarWidthString = getParameter("scrollbar");
            if (scrollbarWidthString != null)
                if (scrollbarWidthString.equals(""))
                    scrollbarWidth = 0;
                else
                    scrollbarWidth = Integer.parseInt(scrollbarWidthString,10);
            if (scrollbarWidth < 6)
                scrollbarWidth = 0;
            else if (scrollbarWidth > 50)
                scrollbarWidth = 50;
        } catch (NumberFormatException e) {
            // don't care
        }
        bgColor = getBackground();
        try {
            String bgColorString = getParameter("bgcolor");
            if (bgColorString != null && bgColorString.startsWith("#")) {
                String hexString = bgColorString.substring(1);
                int bgColorValue = Integer.parseInt(hexString,16);
                if (bgColorValue >= 0 &&
                    bgColorValue <= 0xFFFFFF) {
                    bgColor      = new Color(bgColorValue);
                }
            }
        } catch (NumberFormatException e) {
        }
        setBackground(bgColor);
        if (configurationDialog.getSetting(ConfigurationDialog.VIDEO_MODE_KEY)
            == ConfigurationDialog.UNKNOWN) {
            String videoModeString = getParameter("reverse");
            videoMode = (videoModeString != null);
            configurationDialog.changeSetting(
                                      ConfigurationDialog.VIDEO_MODE_KEY,
                                      videoMode ? ConfigurationDialog.REVERSE :
                                      ConfigurationDialog.NORMAL, false);
        }
        Dimension currentSize = size();
        width  = currentSize.width;
        height = currentSize.height;
        return(true);
    }


    private URL getURL(String parameter,String defaultValue) {
        String urlString = getParameter(parameter);
        if (urlString == null)
            if (defaultValue == null)
                return(null);
            else
                urlString = defaultValue;
        if (urlString.indexOf("://") < 0) {
            String cgiBaseString = getCodeBase().toString();
            if (!cgiBaseString.endsWith("/"))
                cgiBaseString = cgiBaseString + "/";
            urlString = cgiBaseString + urlString;
        }
        try {
            return(new URL(urlString));
        } catch (MalformedURLException e) {
            return(null);
        }
    }



    private void initSimpleGUI() {
        outputArea       = new ConsoleArea(scrollBackLines,
                                        bellType == ConfigurationDialog.VISUAL,
                                        fontURL);
        outputArea.reshape(PANEL_SEP,PANEL_SEP,
                           width-scrollbarWidth-2*PANEL_SEP,
                           height-BUTTON_HEIGHT-2*BUTTON_SEP-2*PANEL_SEP-
                           HEIGHT_FUDGE);
        outputArea.setReverse(videoMode);
        outputArea.setBellType(bellType == ConfigurationDialog.VISUAL);
        Scrollbar outputScrollbar = null;
        if (scrollbarWidth > 0) {
            outputScrollbar  = outputArea.getScrollbar();
            outputScrollbar.reshape(width-scrollbarWidth-PANEL_SEP,
                                    PANEL_SEP,scrollbarWidth,
                                    height-2*PANEL_SEP-HEIGHT_FUDGE);
        }
        outputEmulator   = new VT102(outputArea);
        outputEmulator.setCursorKeyType(cursorKeyType);
        outputEmulator.setBackspaceType(backspaceType);
        inputField       = new TextField();
        inputField.setEditable(true);
        inputField.reshape(PANEL_SEP+BUTTON_SEP,height-BUTTON_HEIGHT-
                           BUTTON_SEP-PANEL_SEP-HEIGHT_FUDGE,width-
                           BUTTON_WIDTH-4*BUTTON_SEP-scrollbarWidth-PANEL_SEP,
                           BUTTON_HEIGHT);
        inputField.setFont(new Font("Serif",Font.PLAIN,14));
        connectionButton = new Button();
        connectionButton.reshape(width-BUTTON_WIDTH-BUTTON_SEP-scrollbarWidth-
                                 PANEL_SEP,height-BUTTON_HEIGHT-BUTTON_SEP-
                                 PANEL_SEP-HEIGHT_FUDGE,BUTTON_WIDTH,
                                 BUTTON_HEIGHT);
        connectionButton.setFont(new Font("Serif",Font.BOLD,14));
        connectionButton.setLabel(CONNECT);
        Panel panel = new AppletPanel();
        panel.setLayout(null);
        panel.reshape(0,0,width,height-HEIGHT_FUDGE);
        panel.add(outputArea);
        panel.add(inputField);
        panel.add(connectionButton);
        if (outputScrollbar != null)
            panel.add(outputScrollbar);
        setLayout(null);
        add(panel);
        repaint();
        return;
    }


    public boolean handleEvent(Event ev) {
        if (consoleIsFocused && ev.target != outputArea &&
            (ev.id == Event.KEY_PRESS || ev.id == Event.KEY_ACTION))
            return(outputArea.handleEvent(ev));
        if (ev.target == inputField &&
            ev.id == Event.ACTION_EVENT) {
            String inputText = inputField.getText();
            inputField.setText("");
            try {
                tunnel.submitText(inputText + "\n");
            } catch (TunnelException e) {
                showErrorMessage(e.getMessage());
            } catch (NullPointerException e) {
                showErrorMessage("Not connected");
            }
            return(true);
        } else if (ev.target == connectionButton &&
                   ev.id == Event.ACTION_EVENT) {
            try {
                if (tunnel == null || !tunnel.isConnected()) {
                    outputEmulator.reset();
                    outputArea.clearScrollBackBuffer();
                    inputField.setText("");
                    outputArea.requestFocus();
                    tunnel = new Tunnel(cgiURL,ShellInABox.this);
                    Dimension charSize  = outputArea.getCharSize();
                    Dimension pixelSize = outputArea.getPixelSize();
                    tunnel.connect(terminalType,backspaceType,
                                   charSize,pixelSize);
                    connectionButton.setLabel(DISCONNECT);
                    showStatus("Connected!");
                } else {
                    tunnel.disconnect();
                }
            } catch (TunnelException e) {
                showErrorMessage(e.getMessage());
            }
            return(true);
        }
        return(super.handleEvent(ev));
    }


    public void setTerminalType(int terminalType) {
        switch (terminalType) {
        case ConfigurationDialog.LINUX:
            this.terminalType = "linux";
            break;
        case ConfigurationDialog.ANSI:
            this.terminalType = "ansi";
            break;
        case ConfigurationDialog.VT102:
            this.terminalType = "vt102";
            break;
        case ConfigurationDialog.XTERM:
            this.terminalType = "xterm";
            break;
        case ConfigurationDialog.XTERM_COLOR:
            this.terminalType = "xterm-color";
            break;
        }
        return;
    }


    public void setCursorKeyType(int cursorKeyType) {
        switch (cursorKeyType) {
        case ConfigurationDialog.LINUX:
            this.cursorKeyType = VT102.LINUX;
            break;
        case ConfigurationDialog.ANSI:
            this.cursorKeyType = VT102.ANSI;
            break;
        case ConfigurationDialog.VT102:
            this.cursorKeyType = VT102.VT102;
            break;
        case ConfigurationDialog.XTERM:
            this.cursorKeyType = VT102.XTERM;
            break;
        case ConfigurationDialog.XTERM_COLOR:
            this.cursorKeyType = VT102.XTERM_COLOR;
            break;
        }
        if (outputEmulator != null)
            outputEmulator.setCursorKeyType(this.cursorKeyType);
        return;
    }


    public void setBackspaceType(int backspaceType) {
        this.backspaceType = backspaceType == ConfigurationDialog.DELETE;
        if (outputEmulator != null)
            outputEmulator.setBackspaceType(this.backspaceType);
        return;
    }


    public void setFontSize(int fontSize) {
        return;
    }


    public void setBellType(int bellType) {
        this.bellType = bellType;
        if (outputArea != null)
            outputArea.setBellType(bellType == ConfigurationDialog.VISUAL);
        return;
    }


    public void setClipboardType(int clipboardType) {
        nativeClipboard = (clipboardType == ConfigurationDialog.NATIVE);
        clipboard = Clipboard.getSystemClipboard();
        clipboard.enableNativeClipboard(nativeClipboard);
        return;
    }


    public void setVideoMode(int videoMode) {
        this.videoMode = videoMode == ConfigurationDialog.REVERSE;
        if (outputArea != null)
            outputArea.setReverse(this.videoMode);
        return;
    }
    

    private class ConsoleArea extends VectorFontConsole {

        public ConsoleArea(int scrollBackLines,boolean visualBell,URL fontURL){
            super(scrollBackLines,visualBell,fontURL);
            return;
        }


        public boolean gotFocus(Event ev,Object what) {
            consoleIsFocused = true;
            return(super.gotFocus(ev,what));
        }


        public boolean lostFocus(Event ev,Object what) {
            consoleIsFocused = false;
            return(super.lostFocus(ev,what));
        }


        public boolean handleEvent(Event ev) {
            String inputText = null;
            if (ev.id == Event.KEY_PRESS) {
                char  key = (char)ev.key;
                if (key == '\177') {
                    ev.id  = Event.KEY_ACTION;
                    ev.key = Event.DELETE;
                } else if (key == '\010') {
                    ev.id  = Event.KEY_ACTION;
                    ev.key = Event.BACK_SPACE;
                } else if (key == '\011') {
                    ev.id  = Event.KEY_ACTION;
                    ev.key = Event.TAB;
                } else if (key == '\000' && ev.modifiers == 0) {
                    // glitch in the VM
                } else {
                    inputText = "";
                    if ((ev.modifiers & Event.CTRL_MASK) != 0)
                        key &= 0x1F;
                    if ((ev.modifiers & (Event.META_MASK|Event.ALT_MASK)) != 0
                        && outputEmulator.getMetaMode())
                        inputText = "\033";
                    inputText = inputText + key;                    
                }
            }
            if (ev.id == Event.KEY_ACTION) {
                inputText = outputEmulator.parseKeyCode(ev.key,ev.modifiers);
            }
            if (ev.target == this &&
                cellWidth != 0 && cellHeight != 0 &&
                (ev.id == Event.MOUSE_DOWN  ||
                 ev.id == Event.MOUSE_UP    ||
                 ev.id == Event.MOUSE_MOVE  ||
                 ev.id == Event.MOUSE_DRAG  ||
                 ev.id == Event.MOUSE_ENTER ||
                 ev.id == Event.MOUSE_EXIT)) {
                int xScreen = (ev.x - (ev.x<0 ? cellWidth-1  : 0))/cellWidth;
                int yScreen = (ev.y - (ev.y<0 ? cellHeight-1 : 0))/cellHeight;
                int yBuffer = yScreen + scrollOffset;
                int button  = (ev.id == Event.MOUSE_DOWN ||
                               ev.id == Event.MOUSE_UP   ||
                               ev.id == Event.MOUSE_DRAG) ?
                              (ev.modifiers & Event.META_MASK ) != 0 ? 4 :
                              (ev.modifiers & Event.ALT_MASK) != 0 ? 2 : 1 : 0;
                int modifier=(ev.id == Event.MOUSE_DOWN ||
                              ev.id == Event.MOUSE_UP ||
                              ev.id == Event.MOUSE_DRAG) ?
                             ev.modifiers & ~(Event.ALT_MASK|Event.META_MASK):
                             ev.modifiers;
                inputText = super.processMouseEvent(ev.id,xScreen,yBuffer,
                                                    ev.clickCount,button,
                                                    modifier);
                if (inputText == null && ev.id == Event.MOUSE_UP &&
                    xScreen >= 0 && xScreen < charWidth &&
                    yScreen >= 0 && yScreen < charHeight) {
                    if ((button & 4) != 0) {
                        configurationDialog.show();
                    } else {
                        try {
                            String urlString = findURL(
                                        new String(consoleCharBuffer[yBuffer]),
                                        new String(consoleAttrBuffer[yBuffer]),
                                        xScreen);
                            URL url = new URL(urlString);
                            AppletContext context = getAppletContext();
                            context.showDocument(url,"_blank");
                            return(true);
                        } catch (Exception e) {
                            // If anything went wrong, just ignore it...
                        }
                    }
                }
            }
            if (inputText != null && inputText.length() > 0) {
                try {
                    tunnel.submitText(inputText);
                } catch (TunnelException e) {
                    showErrorMessage(e.getMessage());
                } catch (NullPointerException e) {
                    showErrorMessage("Not connected");
                }
                return(true);
            }
            return(super.handleEvent(ev));
        }


        public void beep() {
            if (bellType == ConfigurationDialog.FANCY) {
                try {
                    beepAudioClip.play();
                } catch (Exception e) {
                    super.beep();
                }
            } else {
                super.beep();
            }
            return;
        }


        public void displayStatus(String status) {
            showStatus(status);
            return;
        }


        protected void resetConsole(Graphics g,int width,int height) {
            super.resetConsole(g,width,height);
            try {
                tunnel.disconnect();
            } catch (NullPointerException e) {
                // don't care
            } catch (TunnelException e) {
                // don't care
            }
            String[] msg = {
                "                                       ",
                "        Welcome to ShellInABox         ",
                null,
                "                                       ",
                "       \u00A9 1999 by Markus Gutschke       ",
                "         \u001B[36mhttp://www.ShellInABox.com\u001B[37m    ",
                "                                       ",
                "  The source code for this applet is   ",
                "  available under the conditions of    ",
                "  the GPL Gnu General Public License.  ",
                "                                       "
            };
            width      = msg[0].length();
            height     = msg.length;
            Dimension charSize = getCharSize();
            int xOff   = (charSize.width - width)/2 + 1;
            int yOff   = (charSize.height - height)/2 + 1;
            if (xOff < 3 || yOff < 3)
                return;
            String version = VERSION;
            while (version.length() < width) {
                version = version + " ";
                if (version.length() == width)
                    break;
                version = " " + version;
            }
            msg[2] = version;
            outputEmulator.writeString("\u001B[?25l" +
                                       "\u001B[44;37;1m" +
                                       "\u001B["+(yOff-1)+";"+(xOff-1)+"H" +
                                       "\u250C");
            for (int i = 0; i < width; i++)
                outputEmulator.writeString("\u2500");
            outputEmulator.writeString("\u2510\u001B[m\u001B[21m\u2584" +
                                       "\u001B[44;37;1m");
            for (int i = 0; i < msg.length; i++)
                outputEmulator.writeString("\u001B["+(yOff+i)+";"+(xOff-1)+"H"+
                                           "\u2502" +
                                           msg[i] +
                                           "\u2502" +
                                           "\u001B[m\u001B[21m\u2588" +
                                           "\u001B[44;37;1m");
            outputEmulator.writeString("\u001B["+(yOff+height)+";"+(xOff-1)+"H"
                                      +"\u2514");
            for (int i = 0; i < width; i++)
                outputEmulator.writeString("\u2500");
            outputEmulator.writeString("\u2518" +
                                       "\u001B[m\u001B[21m\u2588" +
                                       "\u001B["+(yOff+height+1)+";"+
                                       (xOff)+"H");
            for (int i = 0; i < width+2; i++)
                outputEmulator.writeString("\u2580");
            showStatus("Press \"Connect\" to start a new session...");
            return;
        }


        private boolean isURLCharacter(char ch) {
            return(ch > ' ' && ch < '\u0080');
        }


        private String findURL(String src,String attr,int idx) {
            char curAttr = attr.charAt(idx);
            while (idx > 0) {
                char ch       = src.charAt(idx-1);
                char nextAttr = attr.charAt(idx-1);
                if (!isURLCharacter(ch) || curAttr != nextAttr)
                    break;
                idx--;
            }
            int l = 0;
            while (idx+l < src.length() &&
                   isURLCharacter(src.charAt(idx+l)) &&
                   curAttr == attr.charAt(idx+l))
                l++;
            if (l <= 0) {
                return(null);
            } else {
                return(src.substring(idx,idx+l));
            }
        }
    }


    private class AppletPanel extends Panel {

        public void paint(Graphics g) {
            Dimension curSize  = size();
            g.setColor(Color.black);
            g.drawRect(0,0,curSize.width-1,curSize.height-1);
            return;
        }
    }
}

/*
 * Local variables:
 *  mode: java
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
