package com.shellinabox.clipboard;

import java.awt.Toolkit;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import netscape.security.PrivilegeManager;


public class Clipboard {

    private static Clipboard sharedInstance = getSystemClipboard();

    protected String         contents;
    protected ClipboardOwner owner;


    public static Clipboard getSystemClipboard() {
        if (sharedInstance == null)
            try {
                Class helperClass = Class.forName(Clipboard.class.getName() +
                                                  "$Helper");
                sharedInstance = (Clipboard)helperClass.newInstance();
            } catch (Throwable e) {
                sharedInstance = new Clipboard();
            }
        return(sharedInstance);
    }


    public void setContents(String contents,ClipboardOwner owner) {
        synchronized (this) {
            if (this.owner != null)
                this.owner.lostOwnership(this,this.contents);
            this.contents    = contents;
            this.owner       = (contents != null) ? owner : null;
        }
        return;
    }


    public String getContents(Object requestor) {
        return(contents);
    }


    protected Clipboard() {
        return;
    }


    public void enableNativeClipboard(boolean useNativeClipboard) {
        return;
    }


    /**
     * Clutch to dynamically load and access java.awt.datatransfer. This is
     * surprisingly difficult to get right, because the original Java
     * bytecode verifier was very picky about not loading classes if
     * they depended on other non-existant class files.
     */
    static class Helper extends Clipboard
                        implements java.awt.datatransfer.ClipboardOwner,
                                   Transferable {
        private static final String CLIPBOARDACCESS =
                                              "UniversalSystemClipboardAccess";

        private PrivilegeManager                privilegeManager;
        private String                          contents;
        private ClipboardOwner                  owner;
        private java.awt.datatransfer.Clipboard systemClipboard;
        private Toolkit                         toolkit;
        private boolean                         useNativeClipboard;


        public Helper() {
            super();
            privilegeManager = PrivilegeManager.getPrivilegeManager();
            privilegeManager.enablePrivilege(CLIPBOARDACCESS);
            toolkit         = Toolkit.getDefaultToolkit();
            systemClipboard = toolkit.getSystemClipboard();
        }


        public void enableNativeClipboard(boolean useNativeClipboard) {
            if (!this.useNativeClipboard || owner != null)
                systemClipboard.setContents(null,null);
            this.useNativeClipboard = useNativeClipboard;
            contents                = null;
            owner                   = null;
            super.contents          = null;
            super.owner             = null;
            return;
        }


        public void setContents(String contents,ClipboardOwner owner) {
            if (useNativeClipboard) {
                synchronized (this) {
                    privilegeManager.enablePrivilege(CLIPBOARDACCESS);
                    systemClipboard.setContents(this,this);
                    this.contents    = contents;
                    this.owner       = (contents != null) ? owner : null;
                }
            } else
                super.setContents(contents,owner);
            return;
        }


        private String getStringFromTransferable(Transferable transferable) {
            if (transferable == null) {
                return(null);
            }
            try {
                String stringTransferable = (String)
                         transferable.getTransferData(DataFlavor.stringFlavor);
                return(stringTransferable);
            } catch (Exception e) {
                // don't care
            }
            try {
                InputStream inputStream = (InputStream)
                      transferable.getTransferData(DataFlavor.plainTextFlavor);
                StringBuffer buffer = new StringBuffer();
                byte[]  tmp         = new byte[1024];
                int     count;
                while ((count = inputStream.read(tmp)) >= 0) {
                    for (int i = 0; i < count; i++) 
                        buffer.append((char)tmp[i]);
                }
                inputStream.close();
                return(buffer.toString());
            } catch (Exception e) {
                // don't care
            }
            return(null);
        }


        public String getContents(Object requestor) {
            if (useNativeClipboard) {
                privilegeManager.enablePrivilege(CLIPBOARDACCESS);
                Transferable contents = systemClipboard.getContents(this);
                return(getStringFromTransferable(contents));
            } else
                return(super.getContents(requestor));
        }


        public void lostOwnership(java.awt.datatransfer.Clipboard clipboard,
                                  Transferable contents) {
            String contentsString = getStringFromTransferable(contents);
            if (owner != null) {
                owner.lostOwnership(this,contentsString);
                owner         = null;
                this.contents = null;
            }
            return;
        }


        public DataFlavor[] getTransferDataFlavors() {
            DataFlavor flavors[] = {DataFlavor.stringFlavor,
                                    DataFlavor.plainTextFlavor};
            return(flavors);
        }


        public boolean isDataFlavorSupported(DataFlavor flavor) {
            return(flavor.equals(DataFlavor.stringFlavor) ||
                   flavor.equals(DataFlavor.plainTextFlavor));

        }


        public Object getTransferData(DataFlavor flavor)
                               throws UnsupportedFlavorException, IOException {
            if (contents == null) {
                return(null);
            }
            if (flavor.equals(DataFlavor.stringFlavor))
                return(contents);
            else if (flavor.equals(DataFlavor.plainTextFlavor))
                return new StringReader(contents);
            else
                throw new UnsupportedFlavorException(flavor);
        }
    }
}

/*
 * Local variables:
 * mode: java
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
