package com.shellinabox.tunnel;

import java.awt.Dimension;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.net.URLConnection;


class HttpThread implements Runnable, TunnelConstants {

    private URLConnection serverConnection;
    private boolean       isRunning            = true;
    private boolean       isConnected          = false;
    private int           sessionId            = 0;
    private int           inputSequenceNumber  = 0;
    private int           outputSequenceNumber = 0;
    private GuiObject     guiObject            = null;
    private URL           cgiURL;


    HttpThread(URL cgiURL,GuiObject guiObject,String terminal,
               boolean backspaceSendsDelete,Dimension charSize,
               Dimension pixelSize) throws TunnelException {
        this.cgiURL    = cgiURL;
        this.guiObject = guiObject;
        if (!openConnection(terminal,backspaceSendsDelete,charSize,pixelSize))
            throw new TunnelException("Cannot open connection");
        Thread thread  = new Thread(this,"HttpPollThread");
        thread.start();
        return;
    }
    

    public void run() {
        PrepareConnectionThread connectionThread =
             new PrepareConnectionThread(cgiURL,sessionId,inputSequenceNumber);
        while (isRunning) {
            InputStream inputStream = null;
            try {
                serverConnection = connectionThread.getConnection();
                inputStream = serverConnection.getInputStream();
                byte[] bytes = new byte[CGI_DATA.length()];
                if (inputStream.read(bytes) < CGI_DATA.length() ||
                    !(new String(bytes,0)).equals(CGI_DATA))
                    throw new IOException("Protocol error");
                for (;;) {
                    bytes = new byte[16384];
                    int length = inputStream.read(bytes);
                    if (length <= 0)
                        break;
                    try {
                        if (!isRunning) {
                            connectionThread.destroy();
                            return;
                        }
                    } catch (NullPointerException e) {
                        // Don't care
                    }
                    guiObject.receiveDataEvent(new String(bytes,0,0,length));
                }
                inputStream.close();
            } catch (IOException e) {
                if (inputStream != null) {
                    try {
                        inputStream.close();
                    } catch (IOException ex) {
                    }
                }
                if (isRunning) {
                    disconnect();
                }
                connectionThread.destroy();
                return;
            }
        }
    }
    
    
    void disconnect() {
        if (!isRunning || !isConnected)
            return;
        try {
            isRunning   = false;
            isConnected = false;
            try {
                URLConnection serverConnection = cgiURL.openConnection();
                serverConnection.setDoInput(true);
                serverConnection.setDoOutput(true);
                serverConnection.setUseCaches(false);
                serverConnection.setRequestProperty("Content-type",
                                                    WLCONTENTTYPE);
                String buf = TunnelUtility.toNetworkInt(sessionId) +
                             TunnelUtility.toNetworkInt(0) +
                             CGI_DISCONNECT;
                int length = buf.length();
                byte[] bytes = new byte[length];
                buf.getBytes(0,length,bytes,0);
                OutputStream outputStream =
                                        serverConnection.getOutputStream();
                outputStream.write(bytes);
                outputStream.close();
                InputStream inputStream =serverConnection.getInputStream();
                inputStream.close();
            } catch (IOException e) {
                // Don't care
            }
            synchronized (this) {
                notifyAll();
            }
            guiObject.disconnectEvent();
        } catch (NullPointerException e) {
            // Don't care
        }
        return;
    }


    SessionObject getSessionObject() throws IOException {
        synchronized (this) {
            for (;;) {
                if (!isRunning)
                    throw new IOException();
                if (isConnected) {
                    SessionObject sessionObject =
                        new SessionObject(sessionId,
                                          outputSequenceNumber++);
                    return(sessionObject);
                }
                try {
                    wait();
                } catch (InterruptedException e) {
                    throw new IOException();
                }
            }
        }
    }
    

    private boolean openConnection(String terminal,
                                   boolean backspaceSendsDelete,
                                   Dimension charSize,
                                   Dimension pixelSize) {
        try {
            serverConnection = cgiURL.openConnection();
            serverConnection.setDoInput(true);
            serverConnection.setDoOutput(true);
            serverConnection.setUseCaches(false);
            serverConnection.setRequestProperty("Content-type",
                                                WLCONTENTTYPE);
            String buf = "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\u0000"
                         + CGI_CONNECT
                         + TunnelUtility.toNetworkInt(charSize.width)
                         + TunnelUtility.toNetworkInt(charSize.height)
                         + TunnelUtility.toNetworkInt(pixelSize.width)
                         + TunnelUtility.toNetworkInt(pixelSize.height)
                         + (backspaceSendsDelete ? "\177" : "\010")
                         + terminal + '\u0000';
            int length = buf.length();
            byte[] bytes = new byte[length];
            buf.getBytes(0,length,bytes,0);
            OutputStream outputStream = serverConnection.getOutputStream();
            outputStream.write(bytes);
            outputStream.close();
            InputStream inputStream = serverConnection.getInputStream();
            bytes = new byte[CGI_CONNECT.length()];
            if (inputStream.read(bytes) != CGI_CONNECT.length() ||
                !(new String(bytes,0)).equals(CGI_CONNECT)) {
                inputStream.close();
                throw new IOException();
            }
            bytes = new byte[8];
            if (inputStream.read(bytes) != 8)
                throw new IOException();
            sessionId            = TunnelUtility.fromNetworkInt(bytes,0);
            inputSequenceNumber  =
            outputSequenceNumber = TunnelUtility.fromNetworkInt(bytes,4);
            synchronized (this) {
                isConnected = true;
                this.notifyAll();
            }
            inputStream.close();
        } catch (IOException e) {
            if (isRunning)
                disconnect();
            return(false);
        }
        return(true);
    }        
}

/*
 * Local variables:
 *  mode: java
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
