package com.shellinabox.tunnel;

import java.io.IOException;
import java.io.OutputStream;
import java.net.URL;
import java.net.URLConnection;


class PrepareConnectionThread implements Runnable, TunnelConstants {

    private URLConnection serverConnection = null;
    private boolean       isRunning        = true;
    private int           sessionId;
    private int           sequenceNumber;
    private URL           cgiURL;


    public PrepareConnectionThread(URL cgiURL,int sessionId,
                                   int sequenceNumber) {
        this.cgiURL         = cgiURL;
        this.sessionId      = sessionId;
        this.sequenceNumber = sequenceNumber;
        Thread thread = new Thread(this,"PrepareConnectionThread");
        thread.start();
        return;
    }
    
    
    public URLConnection getConnection() throws IOException {
        synchronized (this) {
            for (;;) {
                if (serverConnection != null) {
                    URLConnection rc = serverConnection;
                    serverConnection = null;
                    notifyAll();
                    return(rc);
                }
                try {
                    wait();
                } catch (InterruptedException e) {
                }
                if (!isRunning)
                    throw new IOException();
            }
        }
    }

    
    public void run() {
        URLConnection newConnection = null;
        for (;;) {
            if (newConnection == null) {
                try {
                    newConnection = cgiURL.openConnection();
                    newConnection.setDoInput(true);
                    newConnection.setDoOutput(true);
                    newConnection.setUseCaches(false);
                    newConnection.setRequestProperty("Content-type",
                                                     WLCONTENTTYPE);
                    String buf = TunnelUtility.toNetworkInt(sessionId) +
                        TunnelUtility.toNetworkInt(++sequenceNumber) +
                        CGI_POLL;
                    int length = buf.length();
                    byte[] bytes = new byte[length];
                    buf.getBytes(0,length,bytes,0);
                    OutputStream outputStream = newConnection.
                        getOutputStream();
                    outputStream.write(bytes);
                    outputStream.close();
                } catch (IOException e) {
                    isRunning = false;
                }
            }
            synchronized (this) {
                if (serverConnection == null) {
                    serverConnection = newConnection;
                    newConnection    = null;
                    notifyAll();
                    continue;
                }
                try {
                    wait();
                } catch (InterruptedException e) {
                }
                if (!isRunning) {
                    return;
                }
            }
        }
    }


    public void destroy() {
        synchronized (this) {
            isRunning = false;
            notifyAll();
        }
        return;
    }
}

/*
 * Local variables:
 *  mode: java
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
