package com.shellinabox.tunnel;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.net.URLConnection;


class SubmitDataThread implements Runnable, TunnelConstants {

    private static final int TIMEOUT = 60;

    private boolean    isRunning     = true;
    private String     data          = null;
    private URL        cgiURL;
    private HttpThread readThread;


    public SubmitDataThread(URL cgiURL,HttpThread readThread) {
        this.cgiURL     = cgiURL;
        this.readThread = readThread;
        Thread thread = new Thread(this,"SubmitDataThread");
        thread.start();
        return;
    }


    public void submit(String data) throws TunnelException {
        synchronized (this) {
            if (!isRunning)
                throw new TunnelException("Connection closed...");
            if (this.data == null)
                this.data = data;
            else
                this.data = this.data + data;
            notifyAll();
        }
        return;
    }


    public void disconnect() {
        synchronized (this) {
            isRunning = false;
            notifyAll();
        }
        return;
    }


    public void run() {
        for (;;) {
            String        data;
            SessionObject sessionObject;
            URLConnection serverConnection;
            OutputStream  outputStream;
            try {
                sessionObject    = readThread.getSessionObject();
                serverConnection = cgiURL.openConnection();
                serverConnection.setDoInput(true);
                serverConnection.setDoOutput(true);
                serverConnection.setUseCaches(false);
                serverConnection.setRequestProperty("Content-type",
                                                    WLCONTENTTYPE);
                outputStream = serverConnection.getOutputStream();
            } catch (IOException e) {
                isRunning = false;
                return;
            } catch (NullPointerException e) {
                isRunning = false;
                return;
            }
            synchronized (this) {
                if (!isRunning)
                    return;
                data      = this.data;
                this.data = null;
                if (data == null) {
                    try {
                        wait(TIMEOUT*1000);
                    } catch (InterruptedException e) {
                    }
                }
            }
            if (data == null)
                data = "";
            String request =
                TunnelUtility.toNetworkInt(sessionObject.sessionId)     +
                TunnelUtility.toNetworkInt(sessionObject.sequenceNumber)+
                CGI_DATA                                                +
                data;
            int length = request.length();
            byte[] bytes = new byte[length];
            request.getBytes(0,length,bytes,0);
            try {
                outputStream.write(bytes);
                outputStream.close();
                InputStream inputStream = serverConnection.getInputStream();
                inputStream.close();
            } catch (IOException e) {
                isRunning = false;
                return;
            } catch (NullPointerException e) {
                isRunning = false;
                return;
            }
        }
    }

}

/*
 * Local variables:
 *  mode: java
 *  c-basic-offset: 4
 *  indent-tabs-mode: nil
 * End:
 */
